// define this to use timer B to do preemptive multitasking
// Warning: the DC debugger will not like this! Only use in code
// run without the debugger.
// Any well written program will not need preemption. Also, preemption
// adds complexity and also will cause problems with functions that are
// not recursive. Also causes shared variable issues.
// It is only included here to test and is not enabled by default.
// The timer B lib can be found in the free libs page: http://shdesigns.org/rabbit/freelibs.html
//#define USE_TIMERB

// define this for debugging messages of task starts/exits
#define COEXEC_VERBOSE
// be sure to add co_exec.lib to the LIB.DIR file in your DC compiler directory.
#use "co_exec.lib"
// size of a global stack
#define TASK_STACK 1024
// size of xmem stack per task, can be up to 4096 bytes
// they do not all need to be  the same size, youc an save xmem on
// tasks that need little stack.
// Note any task that make snetwork calls should have a large stack
// This is especially true if DHCP is used. With DHCP, any task
// that happens to call tcp_tick() (or a net function that uses it)
// may need 3k of stack to handle the DHCP messages.
// normally use 4096 for all tassks that use network calls.
#define STACK_SIZE 2048  // use division of 4096 for best use.

#ifdef USE_TIMERB
#use "timerb.lib"
#endif

#define NUM_TASKS 10
TASK tasks[NUM_TASKS];

unsigned long task_stats[NUM_TASKS];

#ifdef USE_TIMERB
//
// this will do a task switch on a timer interrupt.
//
root void TimerBRoutine(void)
{
#if DEBUG_RST
	if (dkStatusFlags&0x4)	// do task switch only if running
#endif
		yield;
}
#endif
//
// simple keyboard task just echoes keys to output
//
int kbdtask(int param)
{
    int c; // gets rid of warning
    c=0;
    printf("Keyboard task startup\n");
    while (c!=27)
    {
        while (!kbhit())
        	{
				task_stats[param]++;
				sleep(10);
			}
         c=getchar();
        printf("Pressed '%c'\n",c);
    }
    return 0;
}

//
// Fast task, this task yields often so it is thread freindly
//
// this is always task1, so we optimize so it does nothing
// but switch. This is used to estimate task switch time.
//
int  fast_task(int param)
{
    printf("Fast task startup,param=%d\n",param);
    for (;;)
    {
	    yield;
		task_stats[1]++;     // this is always task 1.
	}
}

//
// task that waits for an event
// main task stes event once every 15 seconds
//
char event1;

int event_task(int param)
{
	while (1)
	{
  		clr_event(&event1);
        if (wait_event(&event1,0xffffffffl))
			printf("Event Task %d wake on event.\n",param);
		task_stats[param]++;
 	}
}
//
// this task does not yeild very often. So other tasks have to wait.
// This typically yoieds about once a second. So all tasks will show
// a loop count incrementing the same.
//
// Once TIMERB is enabled, other tasks will run more often.
//
// This task is aborted once avery 20 seconds or so. When this task is not
// running, you can see much better response times in the other tasks.
//
// Obviously, putting a yield; in the inner for loop would, make it
// more "friendly". But this demo shows how a bad task can effict things
// and how timerB can be used for premption.
//
int  slow_task(int param)
{
	unsigned i;
   long	x;
   x=1;
    printf("Slow task startup, param=%d\n",param);
    while (!check_abort())
    {
    		// do noise generator
    		for (i=0;i<20000;i++)
         {
         	if (x&0x80000000l)
            	x=(x<<1)^0x10208001l;
            else
            	x<<=1;
         }
			task_stats[param]++;
#ifndef USE_TIMERB
         yield;
#endif
    }
    return 2222;
}

//
// generic task that wakes up periodically, multiple copies of this are run
//
int  task(int param)
{
	int i;
   i=0;
    printf("Task %d start.\n",param);
    while (!check_abort())
    {
		task_stats[param]++;
        sleep(param*1000l);
		if (++i==10)
		{
			printf("Hello from task: %d\n",param);
			i=0;
		}
    }
    return 3333;
}

char inBuffer[32],outBuffer[1024];

char local_stack[TASK_STACK];

//
// count how many times a task yeilds, requires task updates the task_stats
//
void time_switches(void)
{
	unsigned long start;
	sleep(1); // sync with timer
	start=task_stats[1];
	sleep(1000l); // 1 second time
	start=task_stats[1]-start;
    printf("Task switches: %5ld/sec\n",start);
}

void main(void)
{
	unsigned char c1;
    int i,zzz;
	int slow_id,odd;
   odd=zzz=0;
	memset(task_stats,0,sizeof(task_stats));
	//
	// if running in flash, use serial portfor stdio
	//
    printf("Startup.\n");
	//
	// FAR_STACKS: Separate Stack segment for each thread. Saves root space.
	// link with coexec_fs.lib
	// init_tasks() must be called before any xallocs so it get 4k alighed stacks.
	// also eact task must allocate stacks that do not cross 4k boundaries. I.e.
	// task 1=2048, task 2=1024, task 3=1024, task 5= any
	// max size is 4k per thread.
	//
    if (init_tasks(NUM_TASKS)) // number of tasks and size to allocate for all stacks
	{
		printf("init tasks failed!");
		return;
	}
	// debug info, should show new DATASEG and SEGSIZE if far stacks used
	printf("SEGSIZE=0x%02x, STACKSEG=0x%02x, DATASEG=0x%02x\n",RdPortI(SEGSIZE),RdPortI(STACKSEG),RdPortI(DATASEG));
	i=1; // used for param, this demo sends a different param to each thread.
	// create tasks, pass  stack, size and param
    create_task(fast_task,/*NULL*/local_stack,sizeof(local_stack),i++);
	// test task switching speed
	time_switches();

   task_stats[1]=0;	// reset count
	// create rest of threads,
	// use NULL to use far stacks (FAR_STACKS defined and co_exec_fs.lib)
	// or allocate from main stack (co_exec_.lib)
    create_task(task,NULL,STACK_SIZE,i++);
    create_task(task,NULL,STACK_SIZE,i++);
    create_task(task,NULL,STACK_SIZE,i++);
    slow_id=create_task(slow_task,NULL,STACK_SIZE,i++);
    create_task(event_task,NULL,STACK_SIZE,i++);
    create_task(kbdtask,NULL,STACK_SIZE,i++);
   //
#ifdef USE_TIMERB
	// start preemptive task switch, once every 20ms
   TimerBInit(20000l,1);
#endif
    while (1)
    {
            sleep(1000l);
			task_stats[0]++;
            if (++zzz==20)
            {
                zzz=0;
 			// demo of aborting and restarting a task.
				// abort requires the task to call check_abort()
                if (++odd&1)
                {
                    task_abort(slow_id);
                }
                else
				// only restart if you know the task has exited
                    restart_task(slow_id,slow_id);
            }
            if (zzz==19)
            {
			    // task stacks are filled with 0xff's. The stack check
			    // allows you to see    how much is used.
                printf("Stack checks:\n");
               for (c1=0;c1<task_count;c1++)
                   printf("\t%d bytes of %d in task %d.\n",check_stack(c1),tasks[c1].stack_size,c1);
            }
				// at 15 seconds wake up the event task.
            if (zzz==15)
					set_event(&event1);
			//
			// in this demo, task_stats[] is incremented by each task to
			// count how many times they yield. Previous versions had this in
			// the exec, but was removed to speed up task switching and reduce mem needs.
			//
            printf("Yields:");
            for (c1=0;c1<task_count;c1++)
                printf("\t%ld",task_stats[c1]);
            putchar('\n');
    }

}